/* Orbit.js
 * Copyright (c) Graham Relf, UK, 2006-2016 */ 

/* global Maths */

/**  Gaussian gravitational constant (radians). */
var K = 0.01720209895;

/**  G x mass of sun - assume other masses are negligible. */
var MU = K * K;

function Orbit (name, radius_px, colour, trailLength, 
                epochJD, meanLongDegs, periLongDegs, omegaDegs, iDegs, e, aAU)
{
    this.name = name;
    this.radius_px = radius_px;
    this.colour = colour;
    this.trailIndex = 0;
    this.trailLength = trailLength;
    this.trail1 = new Array (trailLength);
    this.trail2 = new Array (trailLength);
    this.epochJD = epochJD;
    this.meanLongDegs = meanLongDegs;
    this.periLongDegs = periLongDegs;
    this.omegaDegs = omegaDegs;
    this.iDegs = iDegs;
    this.e = e;
    this.aAU = aAU;
    
    this.cosOmega = Maths.cosDegs (omegaDegs);
    this.sinOmega = Maths.sinDegs (omegaDegs);
    this.cosI = Maths.cosDegs (iDegs);
    this.sinI = Maths.sinDegs (iDegs);
    this.sqrt1_ee = Math.sqrt (1.0 - this.e * this.e);
    this.qAU = this.aAU * (1 - this.e);
    this.nRadPerDay = K / Math.sqrt (this.aAU * this.aAU * this.aAU);
    this.wRad = Maths.DEG2RAD * (Maths.in360 (this.periLongDegs - this.omegaDegs));
    this.meanAnomalyDegs = Maths.in360 (this.meanLongDegs - this.periLongDegs);
    this.tauJD = this.epochJD - Maths.DEG2RAD * this.meanAnomalyDegs / this.nRadPerDay;
}

Orbit.prototype.addToTrail1 = function (pt) { this.trail1 [this.trailIndex] = pt; };
Orbit.prototype.addToTrail2 = function (pt) { this.trail2 [this.trailIndex] = pt; };

Orbit.prototype.clearTrails = function ()
{
    for (var i = 0; i < this.trailLength; i++)
    {
        this.trail1 [i] = undefined;
        this.trail2 [i] = undefined;
    }
}

Orbit.prototype.incTrailIndex = function ()
{
    this.trailIndex++; 
    if (this.trailIndex >= this.trailLength) { this.trailIndex = 0; }
}

Orbit.prototype.getName = function () { return this.name; };
Orbit.prototype.getRadiusPx = function () { return this.radius_px; };
Orbit.prototype.getColour = function () { return this.colour; };
Orbit.prototype.getEpochJD = function () { return this.epochJD; };

/** Get semi-major axis of orbit (au). */
Orbit.prototype.getAAU = function () { return this.aAU; };

/** Get eccentricity. */
Orbit.prototype.getE = function () { return this.e; };

/** Get orbital inclination to the ecliptic (in degrees). */
Orbit.prototype.getIDegs = function () { return this.iDegs; };

/** Get mean longitude at the epoch (in degrees). */
Orbit.prototype.getMeanLongDegs = function () { return this.meanLongDegs; };

/** Get longitude of the ascending node (in degrees). */
Orbit.prototype.getOmegaDegs = function () { return this.omegaDegs; };

/** Get longitude of perihelion (in degrees). */
Orbit.prototype.getPeriLongDegs = function () { return this.periLongDegs; };

/** Get perihelion distance (au). */
Orbit.prototype.getQAU = function () { return this.qAU; };

/** Solve Kepler's equation to get the eccentric anomaly (_E) from the
  *  mean anomaly (_M) and eccentricity (this.e). Angles in radians. */
Orbit.prototype.solveKepler = function (_M)
{
    var THRESHOLD = 1.0e-8;
    var isNegative = false;

    if (_M == 0.0) { return 0.0; }

    if (_M < 0.0)
    {
        _M = -_M;
        isNegative = true;
    }

    var estimate = _M;
    var error;
    var abs1e = Math.abs (1.0 - this.e);
    var threshold = THRESHOLD * abs1e;

    if (this.e > 0.8 && _M < Math.PI / 3.0 || this.e > 1.0) // up to 60 degrees
    {
        var trial = _M / abs1e;

        if (trial * trial > 6.0 * abs1e) // cubic term is dominant
        {
            if (_M < Math.PI)
            {
                trial = Math.cbrt (6.0 * _M);
            }
            else // hyperbolic, 5th & higher-order terms predominant
            {
                trial = Maths.asinh (_M / this.e);
            }
        }

        estimate = trial;
    }

    if (this.e < 1.0)
    {
        error = estimate - this.e * Math.sin (estimate) - _M;

        while (Math.abs (error) > threshold)
        {
            estimate -= error / (1.0 - this.e * Math.cos (estimate));
            error = estimate - this.e * Math.sin (estimate) - _M;
        }
    }
    else
    {
        error = this.e * Math.sinh (estimate) - estimate - _M;

        while (Math.abs (error) > threshold)
        {
            estimate -= error / (this.e * Math.cosh (estimate) - 1.0);
            error = this.e * Math.sinh (estimate) - estimate - _M;
        }
    }

    if (isNegative) return -estimate;

    return estimate;
}; // solveKepler

/** Get heliocentric cartesian coordinates (au) at given Julian date. */
Orbit.prototype.getHeliocentricPositionAU = function (jd)
{
    var tMinusTau = jd - this.tauJD;
    var _Mrad = this.nRadPerDay * tMinusTau; // Mean anomaly
    var rAU = 1, fRad = 0;

    if (this.e <= 0.0)
    {
        alert ("Invalid e = " + this.e + " for orbit of " + this.name);
    }
    else
    if (this.e < 1.0) // Ellipse
    {
        var _Erad = this.solveKepler (_Mrad); // Eccentric anomaly
        var cosE = Math.cos (_Erad);
        rAU = this.aAU * (1.0 - this.e * cosE);
        var rSinF = this.aAU * this.sqrt1_ee * Math.sin (_Erad);
        var rCosF = this.aAU * (cosE - this.e);
        fRad = Math.atan2 (rSinF, rCosF); // True anomaly
    }
    else
    if (this.e > 1.0) // Hyperbola
    {
        var _F;

        if (_Mrad <= 5.0 * (this.e - 0.5))
        {
            var _Y = Math.sqrt (8.0 - 8.0 / this.e);
            var sinh3w = 3.0 * _Mrad /(_Y * (this.e - 1.0));
            var w = Maths.asinh (sinh3w) / 3.0;
            _F = _Y * Math.sinh (w);
        }
        else
        {
            _F = Math.log (2.0 * _Mrad / this.e);
        }

        rAU = this.aAU * (this.e * Math.cosh (_F) - 1.0);
        fRad = 2.0 * Math.atan2 (Math.sqrt (this.e + 1.0) * 
                     Math.tanh (_F * 0.5),
                             Math.sqrt (this.e - 1.0));
        }
        else if (this.e == 1.0) // Parabola
        {
            var p = 2.0 * this.qAU;
            var nbar = K / Math.sqrt (p * p * p);
            var cot_s = 3.0 * nbar * tMinusTau;
            var cotTheta = Math.cbrt (1.0 / Math.tan (0.5 * Math.atan2 (1.0, cot_s)));
            var cotThetaSq_1 = cotTheta * cotTheta - 1.0;
            var tan_fOn2 = cotThetaSq_1 / cotTheta;
            rAU = qAU * (1.0 + tan_fOn2 * tan_fOn2);
            fRad = 2.0 * Math.atan2 (cotThetaSq_1, cotTheta);
        }

        var wfRad = this.wRad + fRad;
        var cosWF = Math.cos (wfRad);
        var sinWF = Math.sin (wfRad);
        var x = rAU * (this.cosOmega * cosWF - this.sinOmega * sinWF * this.cosI);
        var y = rAU * (this.sinOmega * cosWF + this.cosOmega * sinWF * this.cosI);
        var z = rAU * sinWF * this.sinI;
        return {x:x, y:y, z:z};
}; // getHeliocentricPositionAU
