/* Viewer.js
 * Copyright (c) Graham Relf, UK, 2006-2016 */ 
var epochJD = 2457755.0; // 2017-01-01T12:00:00Z
var pxPerAu = 160.0;
var tiltFactor = 0.4;
var timer_ms = 25;
var showMercury = false;
var showVenus = false;
var showMars = true;
var showJupiter = true;
var showPlanetLines = true;
var SUN_RADIUS_PX = 20;
var SUN_COLOUR = "#ff0";
var paused = false;
var reversed = false;
var g2, wd, ht, ht2, ox, oy1, oy2, startJD, jd;
var xSun1, ySun1, xSun2, ySun2, xEarth1, yEarth1, xEarth2, yEarth2;
var mercury, venus, earth, mars, jupiter;

function initViewer (options, timeStep_ms, trailLength)
{   if (!!document.createElement ("canvas").getContext)
    {
        var canvas = document.getElementById ("canvas");
        g2 = canvas.getContext ("2d");
        wd = canvas.width;
        ht = canvas.height;
        ht2 = ht / 2;
        ox = wd / 2;
        oy1 = ht / 4;
        oy2 = 3 * oy1;
        timer_ms = timeStep_ms;
        showMercury = options.includes ("Me");
        showVenus = options.includes ("V");
        showMars = options.includes ("Ma");
        showJupiter = options.includes ("J");
        showPlanetLines = options.includes ("L");
        
        mercury = new Orbit ("Mercury", 5, "#ccc", trailLength,
                    epochJD, 105.675, 77.483, 48.310, 7.004, 0.20564, 0.38710);
        venus = new Orbit ("Venus", 8, "#fff", trailLength,
                    epochJD, 51.210, 131.564, 76.633, 3.395, 0.00676, 0.72333);
        earth = new Orbit ("Earth", 10, "#07f", trailLength,
                    epochJD, 101.099, 102.992, 174.832, 0.002, 0.01670, 1.0000);
        mars = new Orbit ("Mars", 5, "#f70", trailLength * 2,
                    epochJD, 9.677, 336.136, 49.508, 1.848, 0.09342,1.5237);
        jupiter = new Orbit ("Jupiter", 15, "#fc6", trailLength * 4, 
                    epochJD, 190.348, 14.368, 100.494, 1.303, 0.04853, 5.2026);

        if (showJupiter) { pxPerAu = 80; }
        
        jd = startJD = dateToJD (new Date ()) - 365; // Start 1 year before now
        draw (); // Calls itself again after timer_ms
    }
}

function draw ()
{
    g2.fillStyle = "#000";
    g2.fillRect (0, 0, wd, ht);
    g2.font = "bold 20px sans-serif";
    g2.fillStyle = "#fff";
    g2.fillText ("GEOCENTRIC", 40, 40);
    g2.fillText ("HELIOCENTRIC", 40, ht - 20);
    var date = new Date (jdToDate(jd));
    g2.fillText (date.toISOString ().substr (0, 10), 40, 10 + ht2);
    drawTrails ();
    var xyzEarth = earth.getHeliocentricPositionAU (jd);
    var ix = Math.round (xyzEarth.x * pxPerAu);
    var iy = Math.round (xyzEarth.y * pxPerAu * tiltFactor);
    xEarth1 = ox;
    yEarth1 = oy1;
    xEarth2 = ox + ix;
    yEarth2 = oy2 - iy;
    xSun1 = ox - ix;
    ySun1 = oy1 + iy;
    xSun2 = ox;
    ySun2 = oy2;
    earth.addToTrail1 (new Point (xSun1, ySun1));
    earth.addToTrail2 (new Point (xEarth2, yEarth2));
    earth.incTrailIndex();

    var colour = earth.getColour();
    drawDisc (xEarth1, yEarth1, earth.getRadiusPx (), colour);
    drawDisc (xEarth2, yEarth2, earth.getRadiusPx (), colour);
    
    drawDisc (xSun1, ySun1, SUN_RADIUS_PX, SUN_COLOUR);
    drawDisc (xSun2, ySun2, SUN_RADIUS_PX, SUN_COLOUR);

    drawLine (xEarth1, yEarth1, xSun1, ySun1, colour);
    drawLine (xEarth2, yEarth2, xSun2, ySun2, colour);
    
    if (showMercury) { drawPlanet (jd, mercury); }    
    if (showVenus) { drawPlanet (jd, venus); }    
    if (showMars) { drawPlanet (jd, mars); }    
    if (showJupiter) { drawPlanet (jd, jupiter); }

    jd = reversed ? jd - 1 : jd + 1;
    
    if (!paused) { setTimeout (draw, timer_ms); }
} // draw

function drawPlanet (jd, planet)
{
    var xyz = planet.getHeliocentricPositionAU (jd);
    var rPx = planet.getRadiusPx ();
    var colour = planet.getColour();
    var ix = Math.round (xyz.x * pxPerAu);
    var iy = Math.round (xyz.y * pxPerAu * tiltFactor);
    var x1 = xSun1 + ix;
    var y1 = ySun1 - iy;
    var x2 = xSun2 + ix;
    var y2 = ySun2 - iy;
    planet.addToTrail1 (new Point (x1, y1));
    planet.addToTrail2 (new Point (x2, y2));
    planet.incTrailIndex ();
    drawDisc (x1, y1, rPx, colour); 
    drawDisc (x2, y2, rPx, colour); 

    if (showPlanetLines)
    {
        drawLine (xEarth1, yEarth1, x1, y1, colour);
        drawLine (xEarth2, yEarth2, x2, y2, colour);
    }
} // drawPlanet

var PI2 = Math.PI * 2;

function drawLine (x1, y1, x2, y2, colour)
{
    g2.beginPath();
    g2.moveTo (x1, y1);
    g2.lineTo (x2, y2);
    g2.lineWidth = 2;
    g2.strokeStyle = colour;
    g2.stroke ();
}

function drawDisc (x, y, radius, colour)
{
    g2.beginPath ();
    g2.arc (x, y, radius, 0, PI2, false);
    g2.closePath ();
    g2.fillStyle = colour;
    g2.fill ();
}

function drawDot (pt, radius, colour)
{
    if (undefined !== pt)
    {
        g2.beginPath ();
        g2.arc (pt.x, pt.y, radius, 0, PI2, false);
        g2.closePath ();
        g2.fillStyle = colour;
        g2.fill ();
    }
}

function drawTrails ()
{
    var colour;
    
    for (var i = 0; i < earth.trailLength; i++)
    {
        drawDot (earth.trail2 [i], 1, earth.colour);
        drawDot (earth.trail1 [i], 1, SUN_COLOUR);
    }

    if (showMercury)
    {
        for (var i = 0; i < mercury.trailLength; i++)
        {
            colour = mercury.getColour ();
            drawDot (mercury.trail1 [i], 1, colour);
            drawDot (mercury.trail2 [i], 1, colour);
        }
    }

    if (showVenus)
    {
        for (var i = 0; i < venus.trailLength; i++)
        {
            colour = venus.getColour ();
            drawDot (venus.trail1 [i], 1, colour);
            drawDot (venus.trail2 [i], 1, colour);
        }
    }

    if (showMars)
    {
        for (var i = 0; i < mars.trailLength; i++)
        {
            colour = mars.getColour ();
            drawDot (mars.trail1 [i], 1, colour);
            drawDot (mars.trail2 [i], 1, colour);
        }
    }

    if (showJupiter)
    {
        for (var i = 0; i < jupiter.trailLength; i++)
        {
            colour = jupiter.getColour ();
            drawDot (jupiter.trail1 [i], 1, colour);
            drawDot (jupiter.trail2 [i], 1, colour);
        }
    }
} // drawTrails

function jdToDate (jd)
{
    /* 1970.0 UT was JD 2440587.5 */
    var msSince1970_0 = (jd - 2440587.5) * 86400000;
    return new Date (msSince1970_0);
}

function dateToJD (date)
{
    var msSince1970_0 = date.getTime ();
    return 2440587.5 + msSince1970_0 / 86400000;
}

function faster () { if (timer_ms > 4) { timer_ms /= 2; } }
function slower () { timer_ms *= 2; }
function reverse () { reversed = !reversed; }

function restart () 
{
    paused = true;
    jd = startJD;
    clearTrails ();
    paused = false;;
}

function clearTrails ()
{
    mercury.clearTrails ();
    venus.clearTrails ();
    earth.clearTrails ();
    mars.clearTrails ();
    jupiter.clearTrails ();
}

function pause ()
{
    paused = !paused;
    
    if (!paused) { draw (); }
}

function tiltUp () 
{
    var wasPaused = paused;
    paused = true;
    clearTrails ();
    if (tiltFactor >= -0.9) { tiltFactor -= 0.1; } 
    paused = wasPaused;
}

function tiltDown () 
{
    var wasPaused = paused;
    paused = true;
    clearTrails ();
    if (tiltFactor <= 0.9) { tiltFactor += 0.1; } 
    paused = wasPaused;
}

function toggleMercury () 
{
    showMercury = !showMercury; 
    if (!showMercury) { mercury.clearTrails (); }
}

function toggleVenus () 
{
    showVenus = !showVenus; 
    if (!showVenus) { venus.clearTrails (); }
}

function toggleMars () 
{
    showMars = !showMars; 
    if (!showMars) { mars.clearTrails (); }
}

function toggleJupiter () 
{
    var wasPaused = paused;
    paused = true;
    showJupiter = !showJupiter; 
    pxPerAu = showJupiter ? 80 : 160;
    clearTrails ();
    paused = wasPaused;
}
